from dataclasses import dataclass

from mlflow import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE


@dataclass
class Permission:
    name: str
    priority: int
    can_read: bool
    can_update: bool
    can_delete: bool
    can_manage: bool


READ = Permission(
    name="READ",
    priority=1,
    can_read=True,
    can_update=False,
    can_delete=False,
    can_manage=False,
)

EDIT = Permission(
    name="EDIT",
    priority=2,
    can_read=True,
    can_update=True,
    can_delete=False,
    can_manage=False,
)

MANAGE = Permission(
    name="MANAGE",
    priority=3,
    can_read=True,
    can_update=True,
    can_delete=True,
    can_manage=True,
)

NO_PERMISSIONS = Permission(
    name="NO_PERMISSIONS",
    priority=100,
    can_read=False,
    can_update=False,
    can_delete=False,
    can_manage=False,
)

ALL_PERMISSIONS = {
    READ.name: READ,
    EDIT.name: EDIT,
    MANAGE.name: MANAGE,
    NO_PERMISSIONS.name: NO_PERMISSIONS,
}


def get_permission(permission: str) -> Permission:
    return ALL_PERMISSIONS[permission]


def _validate_permission(permission: str):
    if permission not in ALL_PERMISSIONS:
        raise MlflowException(
            f"Invalid permission '{permission}'. Valid permissions are: {tuple(ALL_PERMISSIONS)}",
            INVALID_PARAMETER_VALUE,
        )


def compare_permissions(permission1: str, permission2: str) -> bool:
    _validate_permission(permission1)
    _validate_permission(permission2)
    return ALL_PERMISSIONS[permission1].priority <= ALL_PERMISSIONS[permission2].priority



from mlflow.server.handlers import _get_rest_path

HOME = "/"
LOGIN = "/login"
LOGOUT = "/logout"
CALLBACK = "/callback"

STATIC = "/oidc/static/<path:filename>"
UI = "/oidc/ui/<path:filename>"
UI_ROOT = "/oidc/ui/"

# create access token for current user
GET_ACCESS_TOKEN = _get_rest_path("/mlflow/users/access-token")
# get infrmation about current user
GET_CURRENT_USER = _get_rest_path("/mlflow/users/current")
# list of experiments, models, users
GET_EXPERIMENTS = _get_rest_path("/mlflow/experiments")
GET_MODELS = _get_rest_path("/mlflow/registered-models")
GET_USERS = _get_rest_path("/mlflow/users")
# list of experiments, models, filtered by user
GET_USER_EXPERIMENTS = _get_rest_path("/mlflow/users/<string:username>/experiments")
GET_USER_MODELS = _get_rest_path("/mlflow/users/<string:username>/registered-models")
# list of users filtered by experiment, model
GET_EXPERIMENT_USERS = _get_rest_path("/mlflow/experiments/<int:experiment_id>/users")
GET_MODEL_USERS = _get_rest_path("/mlflow/registered-models/<string:model_name>/users")

# CRUD routes from basic_auth
CREATE_USER = _get_rest_path("/mlflow/users/create")
GET_USER = _get_rest_path("/mlflow/users/get")
UPDATE_USER_PASSWORD = _get_rest_path("/mlflow/users/update-password")
UPDATE_USER_ADMIN = _get_rest_path("/mlflow/users/update-admin")
DELETE_USER = _get_rest_path("/mlflow/users/delete")
CREATE_EXPERIMENT_PERMISSION = _get_rest_path("/mlflow/experiments/permissions/create")
GET_EXPERIMENT_PERMISSION = _get_rest_path("/mlflow/experiments/permissions/get")
UPDATE_EXPERIMENT_PERMISSION = _get_rest_path("/mlflow/experiments/permissions/update")
DELETE_EXPERIMENT_PERMISSION = _get_rest_path("/mlflow/experiments/permissions/delete")
CREATE_REGISTERED_MODEL_PERMISSION = _get_rest_path("/mlflow/registered-models/permissions/create")
GET_REGISTERED_MODEL_PERMISSION = _get_rest_path("/mlflow/registered-models/permissions/get")
UPDATE_REGISTERED_MODEL_PERMISSION = _get_rest_path("/mlflow/registered-models/permissions/update")
DELETE_REGISTERED_MODEL_PERMISSION = _get_rest_path("/mlflow/registered-models/permissions/delete")

# manage group permissions
GET_GROUPS = _get_rest_path("/mlflow/groups")
GET_GROUP_USERS = _get_rest_path("/mlflow/groups/<string:group_name>/users")

GET_GROUP_EXPERIMENTS_PERMISSION = _get_rest_path("/mlflow/groups/<string:group_name>/experiments")
CREATE_GROUP_EXPERIMENT_PERMISSION = _get_rest_path("/mlflow/groups/<string:group_name>/experiments/create")
DELETE_GROUP_EXPERIMENT_PERMISSION = _get_rest_path("/mlflow/groups/<string:group_name>/experiments/delete")
UPDATE_GROUP_EXPERIMENT_PERMISSION = _get_rest_path("/mlflow/groups/<string:group_name>/experiments/update")

GET_GROUP_MODELS_PERMISSION = _get_rest_path("/mlflow/groups/<string:group_name>/registered-models")
CREATE_GROUP_MODEL_PERMISSION = _get_rest_path("/mlflow/groups/<string:group_name>/registered-models/create")
DELETE_GROUP_MODEL_PERMISSION = _get_rest_path("/mlflow/groups/<string:group_name>/registered-models/delete")
UPDATE_GROUP_MODEL_PERMISSION = _get_rest_path("/mlflow/groups/<string:group_name>/registered-models/update")



import mlflow
from mlflow_oidc_auth.entities import (
    ExperimentPermission,
    RegisteredModelPermission,
    User,
)
from mlflow_oidc_auth.routes import (
    CREATE_EXPERIMENT_PERMISSION,
    CREATE_REGISTERED_MODEL_PERMISSION,
    CREATE_USER,
    DELETE_EXPERIMENT_PERMISSION,
    DELETE_REGISTERED_MODEL_PERMISSION,
    DELETE_USER,
    GET_EXPERIMENT_PERMISSION,
    GET_REGISTERED_MODEL_PERMISSION,
    GET_USER,
    UPDATE_EXPERIMENT_PERMISSION,
    UPDATE_REGISTERED_MODEL_PERMISSION,
    UPDATE_USER_ADMIN,
    UPDATE_USER_PASSWORD,
)
from mlflow.utils.credentials import get_default_host_creds
from mlflow.utils.rest_utils import http_request_safe


class AuthServiceClient:
    """
    Client of an MLflow Tracking Server that enabled the default basic authentication plugin.
    It is recommended to use :py:func:`mlflow.server.get_app_client()` to instantiate this class.
    See https://mlflow.org/docs/latest/auth.html for more information.
    """

    def __init__(self, tracking_uri: str):
        """
        Args:
            tracking_uri: Address of local or remote tracking server.
        """
        self.tracking_uri = tracking_uri

    def _request(self, endpoint, method, **kwargs):
        host_creds = get_default_host_creds(self.tracking_uri)
        resp = http_request_safe(host_creds, endpoint, method, **kwargs)
        return resp.json()

    def create_user(self, username: str, password: str):
        """
        Create a new user.

        Args:
            username: The username.
            password: The user's password. Must not be empty string.
        """
        # Input validation
        if not username:
            raise ValueError("Username must not be empty.")
        if not password:
            raise ValueError("Password must not be empty.")

        try:
            resp = self._request(
                CREATE_USER,
                "POST",
                json={"username": username, "password": password},
            )
        except mlflow.exceptions.RestException as e:
            raise e

        return User.from_json(resp["user"])

    def get_user(self, username: str):
        """
        Get a user with a specific username.

        Args:
            username: The username.
        """
        if not username:
            raise ValueError("Username must not be empty.")

        try:
            resp = self._request(
                GET_USER,
                "GET",
                params={"username": username},
            )
        except mlflow.exceptions.RestException as e:
            raise e

        return User.from_json(resp["user"])

    def update_user_password(self, username: str, password: str):
        """
        Update the password of a specific user.

        Args:
            username: The username.
            password: The new password.
        """
        if not username:
            raise ValueError("Username must not be empty.")
        if not password:
            raise ValueError("Password must not be empty.")

        try:
            self._request(
                UPDATE_USER_PASSWORD,
                "PATCH",
                json={"username": username, "password": password},
            )
        except mlflow.exceptions.RestException as e:
            raise e

    def update_user_admin(self, username: str, is_admin: bool):
        """
        Update the admin status of a specific user.

        Args:
            username: The username.
            is_admin: The new admin status.
        """
        if not username:
            raise ValueError("Username must not be empty.")
        if not isinstance(is_admin, bool):
            raise ValueError("is_admin must be a boolean value.")

        try:
            self._request(
                UPDATE_USER_ADMIN,
                "PATCH",
                json={"username": username, "is_admin": is_admin},
            )
        except mlflow.exceptions.RestException as e:
            raise e

    def delete_user(self, username: str):
        """
        Delete a specific user.

        Args:
            username: The username.
        """
        if not username:
            raise ValueError("Username must not be empty.")

        try:
            self._request(
                DELETE_USER,
                "DELETE",
                json={"username": username},
            )
        except mlflow.exceptions.RestException as e:
            raise e

    def create_experiment_permission(self, experiment_id: str, username: str, permission: str):
        """
        Create a permission on an experiment for a user.

        Args:
            experiment_id: The id of the experiment.
            username: The username.
            permission: Permission to grant. Must be one of "READ", "EDIT", "MANAGE" and
                "NO_PERMISSIONS".
        """
        if not experiment_id:
            raise ValueError("Experiment ID must not be empty.")
        if not username:
            raise ValueError("Username must not be empty.")
        if permission not in ["READ", "EDIT", "MANAGE", "NO_PERMISSIONS"]:
            raise ValueError("Permission must be one of 'READ', 'EDIT', 'MANAGE', or 'NO_PERMISSIONS'.")

        try:
            resp = self._request(
                CREATE_EXPERIMENT_PERMISSION,
                "POST",
                json={
                    "experiment_id": experiment_id,
                    "username": username,
                    "permission": permission,
                },
            )
        except mlflow.exceptions.RestException as e:
            raise e

        return ExperimentPermission.from_json(resp["experiment_permission"])

    def get_experiment_permission(self, experiment_id: str, username: str):
        """
        Get an experiment permission for a user.

        Args:
            experiment_id: The id of the experiment.
            username: The username.
        """
        if not experiment_id:
            raise ValueError("Experiment ID must not be empty.")
        if not username:
            raise ValueError("Username must not be empty.")

        try:
            resp = self._request(
                GET_EXPERIMENT_PERMISSION,
                "GET",
                params={"experiment_id": experiment_id, "username": username},
            )
        except mlflow.exceptions.RestException as e:
            # Add more specific error handling here if needed
            raise e

        return ExperimentPermission.from_json(resp["experiment_permission"])

    def update_experiment_permission(self, experiment_id: str, username: str, permission: str):
        """
        Update an existing experiment permission for a user.

        Args:
            experiment_id: The id of the experiment.
            username: The username.
            permission: New permission to grant. Must be one of "READ", "EDIT", "MANAGE" and
                "NO_PERMISSIONS".
        """
        if not experiment_id:
            raise ValueError("Experiment ID must not be empty.")
        if not username:
            raise ValueError("Username must not be empty.")
        if permission not in ["READ", "EDIT", "MANAGE", "NO_PERMISSIONS"]:
            raise ValueError("Permission must be one of 'READ', 'EDIT', 'MANAGE', or 'NO_PERMISSIONS'.")

        try:
            self._request(
                UPDATE_EXPERIMENT_PERMISSION,
                "PATCH",
                json={
                    "experiment_id": experiment_id,
                    "username": username,
                    "permission": permission,
                },
            )
        except mlflow.exceptions.RestException as e:
            raise e

    def delete_experiment_permission(self, experiment_id: str, username: str):
        """
        Delete an existing experiment permission for a user.

        Args:
            experiment_id: The id of the experiment.
            username: The username.
        """
        if not experiment_id:
            raise ValueError("Experiment ID must not be empty.")
        if not username:
            raise ValueError("Username must not be empty.")

        try:
            self._request(
                DELETE_EXPERIMENT_PERMISSION,
                "DELETE",
                json={"experiment_id": experiment_id, "username": username},
            )
        except mlflow.exceptions.RestException as e:
            raise e

    def create_registered_model_permission(self, name: str, username: str, permission: str):
        """
        Create a permission on an registered model for a user.

        Args:
            name: The name of the registered model.
            username: The username.
            permission: Permission to grant. Must be one of "READ", "EDIT", "MANAGE" and
                "NO_PERMISSIONS".
        """
        if not name:
            raise ValueError("Name must not be empty.")
        if not username:
            raise ValueError("Username must not be empty.")
        if permission not in ["READ", "EDIT", "MANAGE", "NO_PERMISSIONS"]:
            raise ValueError("Permission must be one of 'READ', 'EDIT', 'MANAGE', or 'NO_PERMISSIONS'.")

        try:
            resp = self._request(
                CREATE_REGISTERED_MODEL_PERMISSION,
                "POST",
                json={"name": name, "username": username, "permission": permission},
            )
        except mlflow.exceptions.RestException as e:
            raise e

        return RegisteredModelPermission.from_json(resp["registered_model_permission"])

    def get_registered_model_permission(self, name: str, username: str):
        """
        Get an registered model permission for a user.

        Args:
            name: The name of the registered model.
            username: The username.
        """
        if not name:
            raise ValueError("Name must not be empty.")
        if not username:
            raise ValueError("Username must not be empty.")

        try:
            resp = self._request(
                GET_REGISTERED_MODEL_PERMISSION,
                "GET",
                params={"name": name, "username": username},
            )
        except mlflow.exceptions.RestException as e:
            raise e

        return RegisteredModelPermission.from_json(resp["registered_model_permission"])

    def update_registered_model_permission(self, name: str, username: str, permission: str):
        """
        Update an existing registered model permission for a user.

        Args:
            name: The name of the registered model.
            username: The username.
            permission: New permission to grant. Must be one of "READ", "EDIT", "MANAGE" and
                "NO_PERMISSIONS".
        """
        if not name:
            raise ValueError("Name must not be empty.")
        if not username:
            raise ValueError("Username must not be empty.")
        if permission not in ["READ", "EDIT", "MANAGE", "NO_PERMISSIONS"]:
            raise ValueError("Permission must be one of 'READ', 'EDIT', 'MANAGE', or 'NO_PERMISSIONS'.")

        try:
            self._request(
                UPDATE_REGISTERED_MODEL_PERMISSION,
                "PATCH",
                json={"name": name, "username": username, "permission": permission},
            )
        except mlflow.exceptions.RestException as e:
            raise e

    def delete_registered_model_permission(self, name: str, username: str):
        """
        Delete an existing registered model permission for a user.

        Args:
            name: The name of the registered model.
            username: The username.
        """
        if not name:
            raise ValueError("Name must not be empty.")
        if not username:
            raise ValueError("Username must not be empty.")

        try:
            self._request(
                DELETE_REGISTERED_MODEL_PERMISSION,
                "DELETE",
                json={"name": name, "username": username},
            )
        except mlflow.exceptions.RestException as e:
            raise e
